/* MAIL2IDX creates index files for all *.TXT mail files
 * found in the present directory and all directories below it.
 * (C) 1993 Johan. K. Reinalda, WG7J
 * Freely usable for non-commercial use only !
 *
 * Syntax:
 * ' mail2idx -d<dirname>' sets mail directory to create index files.
 * Eg: mail2idx -d/nos
 *
 * if called with an argument not starting with - ,
 * this is taken to be a file to be indexed. If this file has
 * no extension, the '.txt' is assumed.
 * Eg: mail2idx wg7j
 *     mail2idx wg7j.txt
 *
 * v1.04 - /spool/areas file is read for message type. (-d option also
 *         works with this)
 * v1.03 - added -? option for help, smtp 'Apparently-to:' handled,
 * v1.02 - bug fix in -d option. (930706)
 * v1.01 - bug fix, no more corrupt index file if message file is empty
 *         or contains no valid messages.  (930629)
 * v1.0  - Initial release, with JNOS 1.10X3 (930622)
 *
 */

#include <io.h>
#include <fcntl.h>
#include <dir.h>
#ifdef MSDOS
#include <dos.h>
#endif
#include "global.h"
#include "mailutil.h"
#include "smtp.h"
#include "files.h"

#define VERSION "JNOS index files creator v1.04\n" \
                "(C) 1993 Johan. K. Reinalda, WG7J\n"

#define fopen fopen
void dirformat __ARGS((char *area));
int lockit __ARGS((char *id));
int unlockit __ARGS((char *id));
char *bgets __ARGS((char * buf,int size,FILE *fp));
void def_index __ARGS((struct mailindex *index));
void readareas __ARGS((char *name,char *mpath));
int is_area __ARGS((char *name));
int isarea;

char idformat[] = "%5.5ld %c %c %-13.13s %-8.8s %-6.6s %6.6s %-35.35s\n";

char *Hdrs[] = {
	"Approved: ",
	"From: ",
	"To: ",
	"Date: ",
	"Message-Id: ",
	"Subject: ",
	"Received: ",
	"Sender: ",
	"Reply-To: ",
	"Status: ",
	"X-BBS-Msg-Type: ",
	"X-Forwarded-To: ",
	"Cc: ",
	"Return-Receipt-To: ",
	"Apparently-To: ",
	"Errors-To: ",
	"Organization: ",
    "Newsgroup: ",
    "Path: ",
    "X-BBS-Hold: ",
    NULLCHAR,
};

int found;

/* Get a line from the text file that was opened in binary mode.
 * Delete the CR/LF at the end !
 */
char *bgets(char * buf,int size,FILE *fp) {
    char *cp;

    if(fgets(buf,size,fp) == NULL)
        return NULL;
    cp = buf;
    while(*cp != '\0') {
        if(*cp == 0xd || *cp == 0xa)
            *cp = '\0';
        cp++;
    }
    return buf;
}

void dirformat(char *area) {

    while(*area) {
        if(*area == '\\')
            *area = '/';
        area++;
    }
}


/* Given a string of the form <user@host>, extract the part inside the
 * angle brackets and return a pointer to it.
 */
char *
getname(cp)
register char *cp;
{
	register char *cp1;

	if ((cp = strchr(cp,'<')) == NULLCHAR)
		return NULLCHAR;
	cp++;	/* cp -> first char of name */
	if ((cp1 = strchr(cp,'>')) == NULLCHAR)
		return NULLCHAR;
	*cp1 = '\0';
	return cp;
}

/* Parse a string in the "Text: Text <user@host>" or "Text: user@host (Text)"
 * formats for the address user@host.
 */
char *
getaddress(string,cont)
char *string;
int cont;		/* true if string is a continued header line */
{
	char *cp, *ap = NULLCHAR;
	int par = 0;

	if((cp = getname(string)) != NULLCHAR) /* Look for <> style address */
		return cp;

	cp = string;
	if(!cont)
		if((cp = strchr(string,':')) == NULLCHAR)	/* Skip the token */
			return NULLCHAR;
		else
			++cp;
	for(; *cp != '\0'; ++cp) {
		if(par && *cp == ')') {
			--par;
			continue;
		}
		if(*cp == '(')		/* Ignore text within parenthesis */
			++par;
		if(par)
			continue;
		if(*cp == ' ' || *cp == '\t' || *cp == '\n' || *cp == ',') {
			if(ap != NULLCHAR)
				break;
			continue;
		}
		if(ap == NULLCHAR)
			ap = cp;
	}
	*cp = '\0';
	return ap;
}


/* return the header type */
int
htype(s)
char *s;
{
	register char *p;
	register int i;

	p = s;
	/* check to see if there is a ':' before and white space */
	while (*p != '\0' && *p != ' ' && *p != ':')
		p++;
	if (*p != ':')
		return NOHEADER;

	for (i = 0; Hdrs[i] != NULLCHAR; i++) {
		if (strnicmp(Hdrs[i],s,strlen(Hdrs[i])) == 0)
			return i;
	}
	return UNKNOWN;
}



/* create mail lockfile */
int
lockit(id)
char *id;
{
    char lockname[FILE_PATH_SIZE];
	int fd;

	/* Try to create the lock file in an atomic operation */
    sprintf(lockname,"%s",id);
    strcpy(&lockname[strlen(id) - 3],"lck");
	if((fd = open(lockname, O_WRONLY|O_EXCL|O_CREAT,0600)) == -1)
		return -1;
    close(fd);
	return 0;
}


/* remove mail lockfile */
int
unlockit(id)
char *id;
{
	char lockname[FILE_PATH_SIZE];

    sprintf(lockname,"%s",id);
    strcpy(&lockname[strlen(id) - 3],"lck");
	return(unlink(lockname));
}

struct areas {
    struct areas *next;
    char *name;
};
struct areas *Areas = NULL;

void readareas(char *name,char *mpath) {
    FILE *fp;
    struct areas *ca,*na;
    char *cp,mp[128],buf[128];

    if((fp=fopen(name,"rt")) == NULL)
        return;

    while(fgets(buf,sizeof(buf),fp) != NULL) {
        if(buf[0] == '#')   /* comment line */
            continue;
        cp = buf;
        /* get first token on line */
        while(*cp) {
            if(*cp == '\n' || *cp == ' ' || *cp == '\t') {
                *cp = '\0';
                break;
            }
            cp++;
        }
        if(buf[0] != '\0') {
            cp = buf;
            while(*cp) {
                if(*cp == '.' || *cp == '\\')
                    *cp = '/';
                cp++;
            }
            sprintf(mp,"%s/%s.txt",mpath,buf);
            na = malloc(sizeof(struct areas));
            na->name = strdup(mp);
            na->next = NULL;
            if(Areas == NULL) {
                Areas = na;
            } else {
                ca->next = na;
            }
            ca = na;
        }
    }
    fclose(fp);
}

int is_area(char *name) {
    struct areas *a;

    a = Areas;
    while(a) {
        if(stricmp(a->name,name) == 0)
            return 1;
        a = a->next;
    }
    return 0;
}

void showareas() {
    struct areas *a;

    a = Areas;
    while(a) {
       printf("%s\n",a->name);
       a = a->next;
    }
}

void def_index(struct mailindex *index) {

    *index->date = '\0';
    *index->from = '\0';
    strcpy(index->subject,"(none)");
    *index->to = '\0';
    if(isarea)
        index->type = 'B';
    else
        index->type = 'P';
    index->status = 'N';
    index->msgid = 0;
}

void index(char *file) {
    FILE *txt,*idx;
    int prev;
    long start,pos;
    struct mailindex index;
    char *cp,*s;
    char len[10];
    char idxname[128];
    char buf[128];


    if(lockit(file)) {
        printf("Can not lock: %s\n",file);
        /* can't get a lock */
        return;
    }
    /* See if this is an area */
    isarea = is_area(file);

    /* It is important that we open the mail files in binary mode,
     * so that we can correctly read unix files that have one byte
     * eol conventions ! (experience is a good teacher ! - WG7J)
     */
    if((txt = fopen(file,"rb")) == NULL)
        return;

    /* Create index file name, and open it */
    found = 1;
    printf("File: %s\n",file);
    strcpy(idxname,file);
    strcpy(&idxname[strlen(file)-3],"idx");
    if((idx = fopen(idxname,"wt")) == NULL) {
        fclose(txt);
        return;
    }

    def_index(&index);
    start = 0;
    pos = 0;
    prev = 0;
    while(bgets(buf,sizeof(buf),txt) != NULLCHAR) {
        /* Find start of message */
        if (!strncmp(buf,"From ",5)) {
            if(prev) {
                sprintf(len,"%ld",pos-start);
                /* rewrite the index file for the previous message */
                fprintf(idx,idformat,index.msgid,index.type,index.status,
                    index.to,index.from,index.date,len,index.subject);
            }
            /* start of next message; at this point
             * pos has the offset of the start of this line
             */
            start = pos;
            def_index(&index);
            /* Read the SMTP header, but first
             * read the 'Received...' and 'ID... lines'
             * to get the msgid - WG7J
             */
            fgets(buf,sizeof(buf),txt);
            fgets(buf,sizeof(buf),txt);
            if((cp=strstr(buf,"AA")) != NULLCHAR)
                /*what follows is the message-number*/
                index.msgid = atol(cp+2);

            while(bgets(buf,sizeof(buf),txt) != NULLCHAR) {
                if(*buf == '\0') {
                    prev = 1;
                    break; /* end of headers */
                }
                switch(htype(buf)) {
                case APPARTO:
                    if(*index.to != '\0')
                        break;
                    /* Notice fall-through */
                case TO:
                    cp = getaddress(buf,0);
                    sprintf(index.to,"%.13s",cp != NULLCHAR ? cp : "");
                    break;
                case STATUS:
                    index.status = buf[8];
                    break;
                case FROM:
                    cp = getaddress(buf,0);
                    sprintf(index.from,"%.8s",
                        cp != NULLCHAR ? cp : "");
                    if((cp=strchr(index.from,'@')) != NULLCHAR)
                        *cp = '\0';         /* get rid of @-host or @-bbs field */
                    break;
                case SUBJECT:
                    cp = &buf[9];
                    while(*cp == ' ')
                        cp++;
                    sprintf(index.subject,"%.35s",cp);
                    break;
                case BBSTYPE:
                    index.type = buf[16];
                    break;
                case DATE:
                    /* find age from ARPA style date */
                    s = &buf[6];
                    while(*s == ' ')
                        s++;
                    /* check to see if there is a "Day, " field */
                    if((cp=strchr(s,',')) != NULLCHAR)  {
                        /* probably standard ARPA style header */
                        cp += 2; /* get past header and DAY field */
                    } else {
                        /* probably a NNTP style message, that has no
                         * "Day, " part in the date line
                         */
                         cp = s;
                    }
                    while(*cp == ' ')
                        cp++;
                    /* now we should be at the start of the
                     * "14 Apr 92 08:14:32" string
                     */
                    if(strlen(cp) < 17)
                        break;
                    if(cp[5] == ' ') {
                        index.date[0] = ' ';
                        strncpy(&index.date[1],cp,5);
                    } else {
                        strncpy(index.date,cp,6);
                        if(index.date[0] == '0')
                            index.date[0] = ' ';
                    }
                    index.date[6] = '\0';
                    break;
                } /* switch */
            } /* while */
        }
        /* Starting position of next line */
        pos=ftell(txt);
    }
    /* rewrite the index file for the last message */
    if(prev) {
        sprintf(len,"%ld",pos - start);
        fprintf(idx,idformat,index.msgid,index.type,index.status,
            index.to,index.from,index.date,len,index.subject);
    }

    fclose(txt);
    fclose(idx);
    unlockit(file);
    return;
}

int checkdir(char *path) {
    char *wildcard,*newpath,*fullname;
    struct ffblk ff;
    int done;

    if((wildcard = malloc(128)) == NULL) {
        puts("can't allocate 'wildcard'");
        return -1;
    }
    if((newpath = malloc(128)) == NULL) {
        puts("Can't allocate 'newpath'");
        free(wildcard);
        return -1;
    }
    if((fullname = malloc(128)) == NULL) {
        puts("Can't allocate 'fullname'");
        free(wildcard);
        free(newpath);
        return -1;
    }

    /* First check all is the files */
    sprintf(wildcard,"%s/*.txt",path);
    done = findfirst(wildcard,&ff,0);
    while(!done){
        sprintf(fullname,"%s/%s",path,ff.ff_name);
        index(fullname);
        done = findnext(&ff);
    }
    /* Now check for sub-directories */
    sprintf(wildcard,"%s/*.*",path);
    done = findfirst(wildcard,&ff,FA_DIREC);
    while(!done){
        if(strcmp(ff.ff_name,".") && strcmp(ff.ff_name,"..")) {
            /* Not the present or 'mother' directory, so create new path */
            sprintf(newpath,"%s/%s",path,ff.ff_name);
            /* Now recurs into it */
            checkdir(newpath);
        }
        done = findnext(&ff);
    }
    free(wildcard);
    free(newpath);
    free(fullname);
    return 0;
}

char HelpTxt[] = "Options:\n"
                 "<filename>   mailfile to index\n"
                 "-d<rootdir>  set new root, same as 'NOS -d<dir>\n"
                 "-?,-h,-H     produce this help.";

void main(int argc,char *argv[]) {
    char *cp;
    int i,len,onefile = 0;
    char path[128];
    char area[128];
    char fn[128];

    puts(VERSION);

    strcpy(path,"/spool/mail");
    strcpy(area,"/spool/areas");

    if(argc > 1) {
        for(i=1;i<argc;i++) {
            dirformat(argv[1]);
            if(argv[i][0] == '-') {
                /* Command line options */
                switch(argv[i][1]) {
                case '?':
                case 'h':
                case 'H':
                    puts(HelpTxt);
                    return;
                case 'd':
                case 'D':
                    /* Set the directory to search and index */
                    cp = &argv[i][2];
                    len=strlen(cp);
                    if(len) {
                        if(cp[len-1] == '/' || cp[len] == '\\')
                            cp[len-1] = '\0';
                        /* Add path in front */
                        strcpy(path,cp);
                        strcpy(area,cp);
                        strcat(path,"/spool/mail");
                        strcat(area,"/spool/areas");
                    }
                    break;
                default:
                    printf("Invalid option %s\n\n",argv[1]);
                    puts(HelpTxt);
                    return;
                }
            } else {
                /* Just one file to do */
                if((cp=strchr(argv[i],'.')) != NULL)
                    *cp = '\0'; /* Cut off extension */
                strcpy(fn,argv[i]);
                strcat(fn,".txt");
                onefile = 1;
            }
        }
    }

    readareas(area,path);

    if(onefile) {
        index(fn);
        if(!found)
            printf("Not found: %s\n", fn);
        return;
    }

    if(checkdir(path))
        puts("ERROR detected, not all index files created!\n");
    if(!found)
        puts("NO files found to index !\n");
}

